//! Blinks an LED
//!
//! This assumes that a LED is connected to pc13 as is the case on the blue pill board.
//!
//! Note: Without additional hardware, PC13 should not be used to drive an LED, see page 5.1.2 of
//! the reference manual for an explanation. This is not an issue on the blue pill.

#![deny(unsafe_code)]
#![no_std]
#![no_main]

use rtt_target::{rprintln, rtt_init_print};
use panic_rtt_target as _;

use stm32f1xx_hal::{
    prelude::*,
    stm32,
    timer::{Timer, Event},
};
use core::sync::atomic::{self, Ordering};

use stm32f1xx_hal as hal;

#[rtic::app(device = stm32f1xx_hal::stm32, peripherals = true)]
const APP: () = {
    // Defining this struct makes shared resources available to tasks; if
    // they can't be statically initialized, they will be initialized by
    // the values returned from `init`
    struct Resources {
        // resources -- these are statically initialized via `init` attributes
        #[init(0)]
        beat: u8,

        // late resources -- these must be initialized in the `init` task and
        // returned in `init::LateResources`
        led1: hal::gpio::gpioc::PC13<hal::gpio::Output<hal::gpio::PushPull>>,
        tmr2: hal::timer::CountDownTimer<stm32::TIM2>,
        tmr3: hal::timer::CountDownTimer<stm32::TIM3>,
    }

    // This task does startup config; the peripherals are passed in thanks to
    // `peripherals = true` in the app definition. They are the `device` and
    // `core` fields of `init::Context`.
    // Any dynamically-configured shared resources in `Resources` must be
    // returned as part of `init::LateResources`.
    #[init]
    fn init(cx: init::Context) -> init::LateResources {
        rtt_init_print!();
        rprintln!("init begin");

        // Set everything to 8MHz using the external clock
        let mut flash = cx.device.FLASH.constrain();
        let mut rcc = cx.device.RCC.constrain();
        let clocks = rcc
            .cfgr
            .use_hse(8.mhz())
            .sysclk(8.mhz())
            .hclk(8.mhz())
            .pclk1(8.mhz())
            .pclk2(8.mhz())
            .adcclk(8.mhz())
            .freeze(&mut flash.acr);

        // LED is on pin C13, configure it for output
        let mut gpioc = cx.device.GPIOC.split(&mut rcc.apb2);
        let led1 = gpioc.pc13.into_push_pull_output(&mut gpioc.crh);

        // Use TIM2 for the beat counter task
        let mut tmr2 = Timer::tim2(cx.device.TIM2, &clocks, &mut rcc.apb1)
            .start_count_down(1.hz());
        tmr2.listen(Event::Update);

        // Use TIM3 for the LED blinker task
        let mut tmr3 = Timer::tim3(cx.device.TIM3, &clocks, &mut rcc.apb1)
            .start_count_down(2.hz());
        tmr3.listen(Event::Update);

        rprintln!("init end");

        init::LateResources {
            led1,
            tmr2,
            tmr3,
        }
    }

    #[idle]
    fn idle(_: idle::Context) -> ! {
        loop {
            // The compiler may omit this loop without the following
            atomic::compiler_fence(Ordering::SeqCst);
        }
    }

    // Update the beat counter and periodically display the current count
    // on the RTT channel
    #[task(resources = [beat])]
    fn beat_update(cx: beat_update::Context) {
        if *cx.resources.beat % 10 == 0 {
            rprintln!("TIM2 beat = {}", *cx.resources.beat);
        }

        *cx.resources.beat += 1;
    }

    // Interrupt task for TIM2, the beat counter timer
    #[task(binds = TIM2, priority = 2, resources = [tmr2], spawn = [beat_update])]
    fn tim2(cx: tim2::Context) {
        // Delegate the state update to a software task
        cx.spawn.beat_update().unwrap();

        // Restart the timer and clear the interrupt flag
        cx.resources.tmr2.start(1.hz());
        cx.resources.tmr2.clear_update_interrupt_flag();
    }

    // Interrupt task for TIM3, the LED blink timer
    #[task(binds = TIM3, priority = 1, resources = [led1, tmr3])]
    fn tim3(cx: tim3::Context) {
        cx.resources.led1.toggle().unwrap();
        cx.resources.tmr3.start(2.hz());
        cx.resources.tmr3.clear_update_interrupt_flag();
    }

    // RTIC requires that unused interrupts are declared in an extern block when
    // using software tasks; these free interrupts will be used to dispatch the
    // software tasks.
    //
    // For a list, see:
    //   https://docs.rs/stm32f1xx-hal/0.6.1/stm32f1xx_hal/stm32/enum.Interrupt.html
    extern "C" {
        fn TAMPER();
    }
};

